""" 
========================================================================
|                                                                      |
|                              SUSY-AI                                 |
|   ----------------------------------------------------------------   |
|                           version 1.1.5                              |
|                          April 7th, 2017                             |
|                                                                      |
|                                                                      |
|   SUSY-AI is a powerful, yet simple interface between SUSY           |
|   data and the sklearn Python class. With the classifier that        |
|   comes with SUSY-AI, it allows for predicting of exclusion of a     |
|   datapoint in pMSSM parameterspace via scikit-learn using a         |
|   pre-trained classifier in only a couple of commands.               |
|                                                                      |
|   Full documentation, and the most recent version of the software    |
|   can be found on the project website                                |
|                                                                      |
|       https://susyai.hepforge.org/                                   |
|                                                                      |
|   If you are using this script scientifically or commercially,       |
|   please reference the following paper                               |
|                                                                      |
|       Sascha Caron, Jong Soo Kim, Krzysztof Rolbiecki, Roberto       |
|       Ruiz de Austri, Bob Stienen, 'The BSM-AI project: SUSY-AI -    |
|       Generalizing LHC limits on Supersymmetry with Machine          |
|       Learning'. arXiv: 1605.02797 [hep-ph]                          |
|                                                                      |
|   For questions about the script, contact                            |
|                                                                      |
|       Bob Stienen <b.stienen@student.ru.nl>                          |
|                                                                      |
========================================================================
"""

__author__ = "Bob Stienen <b.stienen@student.ru.nl>"
__version__ = "1.1.5"


import sys
import os.path
import numpy as np
import time

try:
	import cPickle as pkl
except:
	import pickle as pkl

try:
	from sklearn.ensemble import RandomForestClassifier
except:
	raise ImportError("Python package sklearn has to be installed for SUSY-AI to work.")
	
try:
	import pyslha
except:
	print("")
	print("Warning: SUSY-AI needs the pyslha package to make predictions")
	print("         directly from spectrumfiles, but this package was not found.")
	print("         If this functionality is not used, SUSY-AI will work just fine.")
	print("         It is nevertheless advised to download this package. See")
	print("         documentation for details.")
	print("")


import pyslha


class susyai:
	bins = np.arange(0.005, 1.0, 0.01)
	confidence = [0.99883658, 0.99504328, 0.99291921, 0.98754526, 0.98008268, 0.98230088, 0.97600417, 0.97087379, 0.96936027, 0.9610687, 0.95803609, 0.94995409, 0.9447791, 0.93413487, 0.93352941, 0.92383292, 0.93287756, 0.91268191, 0.91013659, 0.892, 0.897882, 0.87062937, 0.87985866, 0.86776145, 0.85951941, 0.82949309, 0.82436261, 0.8281092, 0.80803134, 0.8142132, 0.79341317, 0.80409042, 0.77389517, 0.76785714, 0.75186368, 0.74463277, 0.71517028, 0.69327731, 0.71763393, 0.69298246, 0.67157895, 0.64756757, 0.61589404, 0.60523446, 0.63675676, 0.61737332, 0.59344262, 0.5685441, 0.55434783, 0.56462585, 0.51653696, 0.50371156, 0.527833, 0.53726708, 0.54713115, 0.56794055, 0.57361377, 0.59512195, 0.61494797, 0.60649087, 0.64510166, 0.66068223, 0.66873339, 0.68049426, 0.70101351, 0.71668138, 0.72285251, 0.74802458, 0.75, 0.77938808, 0.76488095, 0.80615385, 0.79456193, 0.82386364, 0.83476395, 0.84477427, 0.84508816, 0.8627451, 0.86179843, 0.88176241, 0.88331516, 0.90649606, 0.91224591, 0.91837687, 0.91643836, 0.93808312, 0.93674346, 0.93904762, 0.95212766, 0.95958235, 0.95935445, 0.96766938, 0.97106918, 0.97489158, 0.98029256, 0.98485668, 0.99006139, 0.99265971, 0.99499541, 0.9990437]
	mins = [-3999.99, -4000.0, 200.0, 90.879999, 91.419997, 90.02, 90.250001, 200.02999, 102.65, 200.00984, 264.01999, 200.0199, 100.06994, -7998.919, -3999.97, -3999.97, -4381.715, 1364.5547, 2.6724445]
	maxs = [3999.95, 3999.83, 3999.98, 4000.0, 3999.98, 3999.99, 4000.0, 3999.9899, 3999.9999, 3999.98, 3999.9999, 3999.97, 4000.0, 7993.83, 3999.95, 3999.94, 4132.4793, 37569812.0, 66.371989]
	varnames = ['M1', 'M2', 'M3', 'mL1', 'mL3', 'me1', 'me3', 'mQ1', 'mQt', 'mu1', 'mtR', 'md1', 'mbR', 'At', 'Ab', 'Atau', 'mu', 'mA2', 'tanbeta']

	def __init__(self, pickle_location = None, verbosity = 2, coordinate_selector = 1, xtra_selector = None, id_selector = None, map_outsiders = False):
		self.set_classifier()
		self.set_pickle_location(pickle_location)
		self.set_verbosity(verbosity)
		self.set_coordinate_selector(coordinate_selector)
		self.set_xtra_selector(xtra_selector)
		self.set_id_selector(id_selector)
		self.set_map_outsiders(map_outsiders)
		self.set_id_lists()
		self.set_given_warnings()
	
	""" ========================= Getters and setters ========================= """
	# Pickle location
	def get_pickle_location(self):
		return self.pickle_location
	def set_pickle_location(self, pickle_location = None):
		if pickle_location != None and not isinstance(pickle_location, str):
			raise TypeError("Supplied pickle location has to be a string")
		if pickle_location != None and not os.path.isfile(pickle_location):
			raise IOError("Pickle could not be found at indicated location")
		self.pickle_location = pickle_location
	# Verbosity
	def get_verbosity(self):
		return self.verbosity
	def set_verbosity(self, verbosity):
		if isinstance(verbosity, int) and verbosity in [0,1,2]:
			self.verbosity = int(verbosity)
		else:
			raise ValueError("Argument 'verbosity' in set_verbosity() must be in range(3). See documentation for details.")
	# Given warnings
	def get_given_warnings(self, name = None):
		if name == None:
			return self.given_warnings
		else:
			return name in self.get_given_warnings()
	def set_given_warnings(self, name = None):
		if name == None:
			self.given_warnings = []
		else:
			self.given_warnings.append(name)
	# ID lists
	def get_id_lists(self, name = None):
		if name == None:
			return self.id_lists
		else:
			if name in self.get_id_lists():
				return self.id_lists[name]
			else:
				return None
	def set_id_lists(self, name = None, fileid = None):
		if name == None:
			self.id_lists = {}
		else:
			if name not in self.get_id_lists():
				self.id_lists[name] = []
			if fileid != None:
				self.id_lists[name].append(fileid)
	# Coordinate selector
	def get_coordinate_selector(self):
		return self.csel
	def set_coordinate_selector(self, csel = 1):
		if isinstance(csel, int) or isinstance(csel, str):
			if csel == 1 or csel != "1":
				self.csel = [['MSOFT',1],['MSOFT',2],['MSOFT',3],['MSOFT',31],['MSOFT',33],['MSOFT',34],['MSOFT',36],['MSOFT',41],['MSOFT',43],['MSOFT',44],['MSOFT',46],['MSOFT',47],['MSOFT',49],['AU',(3,3)],['AD',(3,3)],['AE',(3,3)],['HMIX',1],['HMIX',4],['HMIX',2]]
			else:
				raise ValueError("Supplied version number for coordinate selection is not valid ('1' currently supported)")
		elif isinstance(csel, int) or isinstance(csel, str):
			self.csel = int(csel)
		elif isinstance(csel, list):
			if len(csel) == 19 and len(csel[0]) == 2:
				self.csel = csel
			else:
				raise ValueError("Supplied coordinate-selector has an unsupported shape, (19,2)-list expected.")
		elif hasattr(csel, '__call__'):
			self.csel = csel
		else:
			raise ValueError("Supplied argument for coordinate-selector is not valid. Supply a supported version number or a method (see documentation for details).")
	# Xtra selector
	def get_xtra_selector(self):
		return self.xsel
	def set_xtra_selector(self, xsel = None):
		if xsel == None or hasattr(xsel, '__call__'):
			self.xsel = xsel
		else:
			raise ValueError("Supplied argument for xtra-selector is not valid. Set it to None or supply a method (see documentation for details).")
	# ID selector
	def get_id_selector(self):
		return self.isel
	def set_id_selector(self, isel = None):
		if isel == None or hasattr(isel, '__call__'):
			self.isel = isel
		elif isinstance(isel, str) and (isel == 'filename' or isel == 'fullfilename'):
			self.isel = isel
		else:
			raise ValueError("Supplied argument for id-selector is not valid. Set it to None, 'filename' or 'fullfilename', or supply a method (see documentation for details).")
	# Map outsiders
	def get_map_outsiders(self):
		return self.map_outsiders
	def set_map_outsiders(self, map_outsiders = False):
			self.map_outsiders = bool(map_outsiders)
	# Classifier
	def get_classifier(self):
		return self.classifier
	def set_classifier(self, classifier = None):
		self.classifier = classifier
		
	""" ========================= Array interfaces ========================= """
	def get_min(self, i):
		if i in range(19):
			return self.mins[i]
		else:
			raise IndexError("get_min("+str(i)+") is out of range: argument must be in range(19)")	
	def get_max(self, i):
		if i in range(19):
			return self.maxs[i]
		else:
			raise IndexError("get_max("+str(i)+") is out of range: argument must be in range(19)")
	def get_name(self, i):
		if i in range(19):
			return self.varnames[i]
		else:
			raise IndexError("get_name("+str(i)+") is out of range: argument must be in range(19)")	

	""" ========================= Messages ========================= """
	def warn(self, text):
		if self.get_verbosity() > 0:
			self.message("\033[93m    WARNING   \033[0m"+str(text))
	def message(self, text):
		if self.get_verbosity() > 1:
			print(str(format(int(round(time.time() * 1000))/1000., '.3f'))+'   '+str(text))
	def warn_once(self, name, text):
		if self.get_verbosity() > 0:
			if not self.get_given_warnings(name):
				self.message("\033[93m    WARNING   \033[0m"+str(text))
				self.set_given_warnings(name)
			
	""" ========================= Pickle loader ========================= """
	def load_pickle(self):
		if self.pickle_location == None:
			raise Exception("No pickle was set to be loaded")
		self.set_classifier(RandomForestClassifier())
		with open(self.pickle_location, 'rb') as f:
			self.message("Loading pickle...")
			self.set_classifier(pkl.load(f))
	
	""" ========================= Select data ========================= """
	def get_spectrum(self, filepath):
		if not os.path.isfile(filepath):
			raise IOError("File '"+filepath+"' could not be found")
		return pyslha.read(filepath)
	@staticmethod
	def dict_to_matrix(dictionary):
		N = int(np.sqrt(len(dictionary)))
		matrix = np.zeros((N, N))
		for i in range(N):
			for j in range(N):
				matrix[i,j] = dictionary[i+1,j+1]
		return matrix
	def select_coordinates(self, spectrum):
		if hasattr(self.get_coordinate_selector(), '__call__'):
			return self.get_coordinate_selector()(spectrum)
		elif isinstance(self.get_coordinate_selector(), list):
			if len(self.get_coordinate_selector()) == 19 and len(self.get_coordinate_selector()[0]) == 2:
				coords = np.zeros(19)
				for i in range(19):
					coords[i] = spectrum.blocks[self.get_coordinate_selector()[i][0]][self.get_coordinate_selector()[i][1]]
				return coords
			else:
				raise ValueError("Supplied coordinate-selector has an unsupported shape, (19,2)-list expected.")
		else:
			raise ValueError("Supplied coordinate-selector was invalid. See documentation for details.")
	def select_xtras(self, spectrum, filepath):
		if self.get_xtra_selector() == None:
			return np.array([])
		elif hasattr(self.get_xtra_selector(), '__call__'):
			return self.get_xtra_selector()(spectrum, filepath)
		else:
			raise ValueError("Supplied xtra-selector was invalid. See documentation for details.")
	def select_id(self, spectrum, filepath):
		if self.get_id_selector() == None:
			return None
		elif hasattr(self.get_id_selector(), '__call__'):
			return self.get_id_selector()(spectrum, filepath)
		elif self.get_id_selector() == 'filename':
			f = filepath.split('/')[-1]
			if '.' in f:
				f = '.'.join(f.split('.')[:-1])
			return f
		elif self.get_id_selector() == 'fullfilename':
			return filepath.split('/')[-1]
		else:
			raise ValueError("Supplied xtra-selector was invalid. See documentation for details.")
	def check_lsp(self, fileid, spectrum):
		blocks = [spectrum.blocks[i].name for i in spectrum.blocks]
		if 'MASS' in blocks:
			masses = np.array([[i, spectrum.blocks['MASS'][i]] for i in spectrum.blocks['MASS'].keys() if i > 1e6])
			masses = masses[abs(masses[:,1]).argsort()]
			if masses[0,0] != 1000022 and 1000022 in masses[:,0]:
				if self.get_id_selector() != None:
					self.warn_once("LSP01", "File detected where ~chi01 is not the LSP. Prediction will continue, but will possibly be incorrect.")
					self.warn_once("LSP02", "All ids of the files for which this holds will be saved internally.")
					self.warn_once("LSP03", "These ids can be retrieved by calling get_id_lists('LSP').")
					self.set_id_lists('LSP',fileid)
					return -1
				else:
					self.warn_once("LSP04", "File detected where ~chi01 is not the LSP. Prediction will continue, but will possibly be incorrect.")
					self.warn_once("LSP05", "No id selector was specified. If it were, all ids of the files for which this holds would be saved internally.")
					self.warn_once("LSP06", "These ids can be retrieved by calling get_id_lists('LSP').")
					return -2
			else:
				return 1
		else:
			self.warn_once("LSP07","LSP could not be identified due to absence of the MASS block or the mass of the lightest neutralino in the spectrumfile.")
			self.warn_once("LSP08","SUSY-AI assumes ~chi01 to be the LSP. Prediction will continue, but is possibly incorrect.")
			return 0
	def get_data_from_file(self, filepath):
		spectrum = self.get_spectrum(filepath)
		coordinates = self.select_coordinates(spectrum)
		xtras = self.select_xtras(spectrum, filepath)
		fileid = self.select_id(spectrum, filepath)
		self.check_lsp(fileid, spectrum)
		return (coordinates, xtras, fileid)
	
	""" ========================= Predict ========================= """
	def map_coordinates(self, coordinates):
		for i in range(19):
			r = self.maxs[i] - self.mins[i]
			newmax = (self.maxs[i] - 0.1125*r)
			newmin = (self.mins[i] + 0.1125*r)
			if len(coordinates.shape) > 1:
				if self.mins[i] < 0:
					coordinates[coordinates[:,i] < newmin, i] = newmin
				else:
					coordinates[coordinates[:,i] < newmin, i] = np.abs(coordinates[coordinates[:,i] < newmin, i])
				coordinates[coordinates[:,i] > newmax, i] = newmax
			else:
				if coordinates[i] < newmin:
					if self.mins[i] < 0:
						coordinates[i] = newmin
					else:
						coordinates[i] = np.abs(coordinates[i])
				if coordinates[i] > newmax:
					coordinates[i] = newmax
		return coordinates
	def calculate_results(self, prediction):
		confidence = np.zeros(len(prediction))
		for i in range(len(prediction)):
			b = np.argmin(abs(self.bins - prediction[i]))
			confidence[i] = self.confidence[b]
		classification = (prediction > self.bins[np.argmin(self.confidence)])*1.0
		return (classification, confidence)
	def apply_min_confidence(self, arrayorlist, confidence, min_confidence):
		if len(arrayorlist) != len(confidence):
			raise ValueError("Lengths of arrayorlist and confidence have to be equal.")
		if type(arrayorlist) != np.ndarray and type(arrayorlist) != list:
			raise ValueError("Inputparameter arrayorlist has to be an array or a list.")
		r = [arrayorlist[i] for i in range(len(arrayorlist)) if confidence[i] >= min_confidence]
		if type(arrayorlist) == np.ndarray:
			r = np.array(r)
		return r
	def predict(self, coordinates, min_confidence = 0.5):
		if min_confidence < 0 or min_confidence > 1.0:
			raise ValueError("Only required certainties between 0 and 1.0 are allowed.")
		if self.get_classifier() == None:
			self.warn("If load_pickle() is not used before using predict(), loading is part of prediction and will therefore be considerably slowed down.")
			self.load_pickle()
		if self.get_map_outsiders() == True:
			self.message("Mapping extreme coordinates...")
			coordinates = self.map_coordinates(coordinates)
		if len(coordinates.shape) == 1:
			coordinates = coordinates.reshape(1, -1)
		self.message("Predicting...")
		prediction = self.get_classifier().predict_proba( coordinates )[:,1]
		self.message("Expanding results...")
		classification, confidence = self.calculate_results(prediction)
		if min_confidence > 0.5:
			self.message("Applying required confidence level...")
			L_orig = len(classification)
			classification = self.apply_min_confidence(classification, confidence, min_confidence)
			prediction = self.apply_min_confidence(prediction, confidence, min_confidence)
			coordinates = self.apply_min_confidence(coordinates, confidence, min_confidence)
			confidence = self.apply_min_confidence(confidence, confidence, min_confidence)
			self.message("   datapoints before: "+str(L_orig))
			self.message("   datapoints after:  "+str(len(confidence)))
		self.message("Prediction finished, result returned...")
		return (classification, prediction, confidence, coordinates)
	
	""" ========================= Predict files ========================= """
	def predict_files(self, filepaths, min_confidence = 0.5):
		N = len(filepaths)
		coordinates = np.zeros((N, 19))
		xtras = None
		fileids = [None] * N
		self.message("Clearing memory of previous runs (if any)...")
		self.set_given_warnings()
		self.set_id_lists()
		self.message("Reading files...")
		for i in range(N):
			c, x, f = self.get_data_from_file(filepaths[i])
			coordinates[i,:] = c
			if self.get_xtra_selector() != None:
				if xtras == None:
					xtras = np.zeros((N,len(x)))
				xtras[i,:] = x
			if self.get_id_selector() != None:
				fileids[i] = f
		classification, prediction, confidence, coordinates = self.predict(coordinates)
		
		if min_confidence > 0.5:
			self.message("Applying required confidence level...")
			L_orig = len(classification)
			classification = self.apply_min_confidence(classification, confidence, min_confidence)
			prediction = self.apply_min_confidence(prediction, confidence, min_confidence)
			coordinates = self.apply_min_confidence(coordinates, confidence, min_confidence)
			if self.get_xtra_selector() != None:
				xtras = self.apply_min_confidence(xtras, confidence, min_confidence)
			if self.get_id_selector() != None:
				fileids = self.apply_min_confidence(fileids, confidence, min_confidence)
			confidence = self.apply_min_confidence(confidence, confidence, min_confidence)
			self.message("   datapoints before: "+str(L_orig))
			self.message("   datapoints after:  "+str(len(confidence)))
		
		result = (classification, prediction, confidence, coordinates)
		if self.get_xtra_selector() != None:
			result = result + (xtras,)
		if self.get_id_selector() != None:
			result = result + (np.array(fileids),)
		return result
